#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>

/* Exercice de colles d'informatique */

/* Titre : Couplage de cardinal maximum en C */

/* On considère dans cet exercice des graphes bipartis. On souhaite
 * proposer une implémentation, en C, de l'algorithme de couplage de
 * cardinal maximum du cours. */


/* Dans toute la suite on suppose qu'il existe une constante entière d_max
 * dans telle que tous les sommets de tous les graphes manipulés ont un
 * degré sortant inférieur à d_max. */

int d_max = 10;

struct graph_s {
    int   nb_sommets ;            /* nombre de sommets du graphe */
    int*  degre      ;            /* tableau des degrés des sommets */
    int** voisins    ;            /* tableau des listes d'adjacence des sommets */
};

typedef struct graph_s* graph;  /* définition du type graph */

/* Les graphes manipulés dans cet exercice sont des graphes orientés, dont
 * les sommets sont des entiers d'un intervalle [|0, n-1|], que l'on
 * représente en C au moyen d'une structure contenant :
 * - un champ entier nb_sommets donnant le nombre de sommets du graphe (n
 ici) ;
 * - un champ degre qui est un tableau de nb_sommets entiers, contenant
 dans sa case d'indice i, le degré sortant du sommet i ;
 * - un champ voisins qui est un tableau de nb_sommets tableaux, contenant
 dans sa case d'indice i un tableau d'entiers de taille d_max, dont les
 degre[i] premières cases contiennent les successeurs du sommet i.
 * On ne spécifie pas ce que contiennent les d_max - degre[i] cases
 * suivantes du tableau voisins[i]. */

/*
 * Fonction : ajoute_arc
 * ----------------------
 * Définir une fonction void ajoute_arc(graph g, int u, int v) qui ajoute l'arc
 * (u, v) au graphe g.
 *
 * Hypothèses :
 * - L'ajout de l'arc ne contrevient pas à la contrainte des degrés sortants :
 *   avant l'ajout, le sommet u a un degré sortant strictement inférieur à d_max.
 * - Le sommet u n'a pas déjà v comme successeur.
 */

/*
 * Fonction : supprime_arc
 * ------------------------
 * Définir une fonction void supprime_arc(graph g, int u, int v) qui supprime
 * l'arc (u, v) du graphe g.
 *
 * Hypothèses :
 * - L'arc (u, v) est bien présent dans le graphe g.
 */

/*
 * Fonction : inverse_arc
 * -----------------------
 * Définir une fonction void inverse_arc(graph g, int u, int v) qui renverse
 * l'arc (u, v) dans le graphe g, c'est-à-dire remplace l'arc (u, v) par (v, u).
 *
 * Hypothèses :
 * - L'arc (u, v) est bien présent dans le graphe g.
 * - Cette inversion respecte la contrainte des degrés sortants, c'est-à-dire
 *   que le sommet v a un degré sortant inférieur à d_max avant l'inversion.
 * - Le sommet v n'a pas déjà u comme successeur.
 */

/*
 * Représentation des chemins dans un graphe :
 * --------------------------------------------
 * Pour un graphe orienté contenant n sommets, un chemin élémentaire est représenté
 * par un tableau de n entiers :
 * - La case d'indice i contient -1 si i n'est pas dans le chemin ou si i est la fin.
 * - Sinon, elle contient j (0 <= j < n) si j est le successeur de i.
 *
 * Exemple :
 * Le chemin 3 -> 4 -> 7 -> 8 dans un graphe de 10 sommets est représenté par :
 * {-1, -1, -1, 4, 7, -1, -1, 8, -1, -1}
 */

typedef int* chemin ;           /* définition d'un chemin dans un graphe */

/*
 * Fonction : trouve_chemin
 * -------------------------
 * Définir une fonction bool trouve_chemin(graph g, int s, int p, chemin* ptr)
 * qui trouve un chemin élémentaire entre deux sommets distincts s et p dans le graphe g.
 *
 * Arguments :
 * - g : le graphe.
 * - s, p : les sommets de départ et d'arrivée.
 * - ptr : un pointeur pour enregistrer l'adresse d'un tableau représentant le chemin.
 *
 * Comportement :
 * - Si un chemin existe, la fonction alloue et remplit le tableau correspondant,
 *   enregistre son adresse dans *ptr et retourne true.
 * - Sinon, elle retourne false et ne modifie pas *ptr.
 *
 * Complexité attendue : linéaire en la taille du graphe.
 */

/*
 * Fonction : inverse_chemin
 * --------------------------
 * Définir une fonction int inverse_chemin(graph g, int s, int p, chemin chm)
 * qui inverse les arcs d'un chemin élémentaire donné entre s et p dans g.
 *
 * Arguments :
 * - g : le graphe.
 * - s, p : sommets de départ et d'arrivée.
 * - chm : chemin élémentaire sous la forme x_0 (=s) -> x_1 -> ... -> x_q (=p).
 *
 * Comportement :
 * - Les arcs (x1, x2), ..., (xq-2, xq-1) sont inversés dans g.
 * - La fonction retourne le dernier sommet avant p (x_(q-1)).
 */

/*
 * Représentation des graphes bipartis au moyen du type graph :
 * -----------------------------------------------------------
 * Les graphes bipartis considérés ici sont des graphes G = (S1 ∪ S2, A) non orientés,
 * où :
 * - S1 = {2, 4, 6, ..., 2m}
 * - S2 = {3, 5, 7, ..., 2m+1}, pour un certain m ∈ ℕ.
 *
 * Ces graphes sont représentés en C au moyen du type `graph` défini précédemment
 * de la manière suivante :
 * - Un sommet 0 est ajouté, relié à tous les sommets x de S1 par un arc orienté (0, x).
 * - Un sommet 1 est ajouté, relié à tous les sommets y de S2 par un arc orienté (y, 1).
 * - Les arêtes de A sont orientées de S1 vers S2.
 *
 * Remarque : Vous n'avez pas besoin d'effectuer cette transformation. Les graphes
 * bipartis fournis à vos fonctions sont déjà représentés sous cette forme.
 */

/*
 * Représentation des couplages :
 * ------------------------------
 * Un couplage C dans un graphe biparti G = (S1 ∪ S2, A), où S1 = {2, 4, 6, ..., 2m}
 * et S2 = {3, 5, 7, ..., 2m+1} pour un certain m ∈ ℕ, est représenté au moyen
 * d'un tableau d'entiers `c` de taille n = 2m+2. Les règles de représentation sont :
 *
 * - c[0] = c[1] = -1;
 * - ∀ i, j ∈ [2, n-1], c[i] = j ⇔ c[j] = i ⇔ {i, j} ∈ C (i et j sont appariés dans le couplage).
 * - ∀ i ∈ [2, n-1], c[i] = -1 ⇔ i est libre dans le couplage C.
 *
 * Le type correspondant à cette représentation est défini dans le code.
 */

typedef int* couplage ;         /* définition d'un couplage */

/*
 * Fonction : couplage_max
 * ------------------------
 * Définir une fonction couplage couplage_max(graph g) qui calcule un couplage
 * de cardinal maximum dans un graphe biparti g.
 *
 * Hypothèses :
 * - Le graphe g est représenté dans la forme orientée décrite, mais peut être modifié.
 *
 * Comportement :
 * - La fonction retourne un couplage maximal sous forme d'un tableau respectant
 *   les conventions de représentation des couplages.
 */

graph cree_graph(int nb_sommets, int d_max) {
  /* crée un graphe contenant nb_sommets sommets, tous de degré sortant
     maximal d_max */
  graph res = (graph) malloc(sizeof(struct graph_s));
  res->nb_sommets = nb_sommets;
  res->degre = (int*) malloc(nb_sommets * sizeof(int));
  res->voisins = (int**) malloc(nb_sommets * sizeof(int*));
  for (int i = 0 ; i < nb_sommets ; i ++) {
    res->degre[i] = 0;
    res->voisins[i] = (int*) malloc(d_max * sizeof(int));
  }
  return res;
}

int main() {
  graph g = cree_graph(10, 13);
  ajoute_arc(g, 0, 2);
  ajoute_arc(g, 0, 4);
  ajoute_arc(g, 0, 6);
  ajoute_arc(g, 0, 8);
  ajoute_arc(g, 3, 1);
  ajoute_arc(g, 5, 1);
  ajoute_arc(g, 7, 1);
  ajoute_arc(g, 9, 1);
  ajoute_arc(g, 2, 3);
  ajoute_arc(g, 2, 5);
  ajoute_arc(g, 4, 3);
  ajoute_arc(g, 4, 7);
  ajoute_arc(g, 6, 5);
  ajoute_arc(g, 8, 5);
  ajoute_arc(g, 8, 7);
  ajoute_arc(g, 8, 9);
  couplage c = couplage_max(g);
  for (int i = 0 ; i < 10 ; i ++) {
    printf("%d, ", c[i]);
  }
}
